public inbox for linux-fsdevel@vger.kernel.org
 help / color / mirror / Atom feed
From: Jan Kara <jack@suse.cz>
To: <linux-fsdevel@vger.kernel.org>
Cc: <linux-block@vger.kernel.org>,
	Christian Brauner <brauner@kernel.org>,
	Al Viro <viro@ZenIV.linux.org.uk>, <linux-ext4@vger.kernel.org>,
	Ted Tso <tytso@mit.edu>,
	"Tigran A. Aivazian" <aivazian.tigran@gmail.com>,
	David Sterba <dsterba@suse.com>,
	OGAWA Hirofumi <hirofumi@mail.parknet.co.jp>,
	Muchun Song <muchun.song@linux.dev>,
	Oscar Salvador <osalvador@suse.de>,
	David Hildenbrand <david@kernel.org>,
	linux-mm@kvack.org, linux-aio@kvack.org,
	Benjamin LaHaise <bcrl@kvack.org>, Jan Kara <jack@suse.cz>
Subject: [PATCH 41/41] fs: Unify generic_file_fsync() with mmb methods
Date: Fri, 20 Mar 2026 14:41:36 +0100	[thread overview]
Message-ID: <20260320134100.20731-82-jack@suse.cz> (raw)
In-Reply-To: <20260320131728.6449-1-jack@suse.cz>

__generic_file_fsync() is practically identical to
generic_mmb_fsync_noflush() with one subtle difference:

1) __generic_file_fsync() takes inode lock when calling writing out the
inode.
2) generic_mmb_fsync_noflush() calls mmb_sync_buffers().

Taking inode lock when writing out the inode seems pointless in
particular because there are lots of places (most notably sync(2) path)
that don't do that so hardly anything can depend on it. When NULL is
passed to generic_mmb_fsync_noflush(), mmb_sync_buffers() is not called
so that difference is not a problem.

So let's remove __generic_file_fsync() and use
generic_mmb_fsync_noflush() instead to reduce code duplication. Arguably
this leaks a bit of buffer_head knowledge into fs/libfs.c which is not
great but avoiding the duplication seems worth it.

Signed-off-by: Jan Kara <jack@suse.cz>
---
 fs/buffer.c                 | 74 -------------------------------------
 fs/exfat/file.c             |  2 +-
 fs/libfs.c                  | 57 ++++++++++++++++------------
 include/linux/buffer_head.h |  5 ---
 include/linux/fs.h          | 12 +++++-
 5 files changed, 45 insertions(+), 105 deletions(-)

diff --git a/fs/buffer.c b/fs/buffer.c
index 43aca5b7969f..591aed740601 100644
--- a/fs/buffer.c
+++ b/fs/buffer.c
@@ -621,80 +621,6 @@ int mmb_sync_buffers(struct mapping_metadata_bhs *mmb)
 }
 EXPORT_SYMBOL(mmb_sync_buffers);
 
-/**
- * generic_mmb_fsync_noflush - generic buffer fsync implementation
- * for simple filesystems with no inode lock
- *
- * @file:	file to synchronize
- * @mmb:	list of metadata bhs to flush
- * @start:	start offset in bytes
- * @end:	end offset in bytes (inclusive)
- * @datasync:	only synchronize essential metadata if true
- *
- * This is a generic implementation of the fsync method for simple
- * filesystems which track all non-inode metadata in the buffers list
- * hanging off the address_space structure.
- */
-int generic_mmb_fsync_noflush(struct file *file,
-			      struct mapping_metadata_bhs *mmb,
-			      loff_t start, loff_t end, bool datasync)
-{
-	struct inode *inode = file->f_mapping->host;
-	int err;
-	int ret = 0;
-
-	err = file_write_and_wait_range(file, start, end);
-	if (err)
-		return err;
-
-	if (mmb)
-		ret = mmb_sync_buffers(mmb);
-	if (!(inode_state_read_once(inode) & I_DIRTY_ALL))
-		goto out;
-	if (datasync && !(inode_state_read_once(inode) & I_DIRTY_DATASYNC))
-		goto out;
-
-	err = sync_inode_metadata(inode, 1);
-	if (ret == 0)
-		ret = err;
-
-out:
-	/* check and advance again to catch errors after syncing out buffers */
-	err = file_check_and_advance_wb_err(file);
-	if (ret == 0)
-		ret = err;
-	return ret;
-}
-EXPORT_SYMBOL(generic_mmb_fsync_noflush);
-
-/**
- * generic_mmb_fsync - generic buffer fsync implementation
- * for simple filesystems with no inode lock
- *
- * @file:	file to synchronize
- * @mmb:	list of metadata bhs to flush
- * @start:	start offset in bytes
- * @end:	end offset in bytes (inclusive)
- * @datasync:	only synchronize essential metadata if true
- *
- * This is a generic implementation of the fsync method for simple
- * filesystems which track all non-inode metadata in the buffers list
- * hanging off the address_space structure. This also makes sure that
- * a device cache flush operation is called at the end.
- */
-int generic_mmb_fsync(struct file *file, struct mapping_metadata_bhs *mmb,
-		      loff_t start, loff_t end, bool datasync)
-{
-	struct inode *inode = file->f_mapping->host;
-	int ret;
-
-	ret = generic_mmb_fsync_noflush(file, mmb, start, end, datasync);
-	if (!ret)
-		ret = blkdev_issue_flush(inode->i_sb->s_bdev);
-	return ret;
-}
-EXPORT_SYMBOL(generic_mmb_fsync);
-
 /*
  * Called when we've recently written block `bblock', and it is known that
  * `bblock' was for a buffer_boundary() buffer.  This means that the block at
diff --git a/fs/exfat/file.c b/fs/exfat/file.c
index 90cd540afeaa..fe6eb391eb4e 100644
--- a/fs/exfat/file.c
+++ b/fs/exfat/file.c
@@ -577,7 +577,7 @@ int exfat_file_fsync(struct file *filp, loff_t start, loff_t end, int datasync)
 	if (unlikely(exfat_forced_shutdown(inode->i_sb)))
 		return -EIO;
 
-	err = __generic_file_fsync(filp, start, end, datasync);
+	err = generic_mmb_fsync_noflush(filp, NULL, start, end, datasync);
 	if (err)
 		return err;
 
diff --git a/include/linux/buffer_head.h b/include/linux/buffer_head.h
index 74fcc9a03c32..f003a1937826 100644
--- a/include/linux/buffer_head.h
+++ b/include/linux/buffer_head.h
@@ -207,11 +207,6 @@ void end_buffer_write_sync(struct buffer_head *bh, int uptodate);
 
 /* Things to do with metadata buffers list */
 void mmb_mark_buffer_dirty(struct buffer_head *bh, struct mapping_metadata_bhs *mmb);
-int generic_mmb_fsync_noflush(struct file *file,
-			      struct mapping_metadata_bhs *mmb,
-			      loff_t start, loff_t end, bool datasync);
-int generic_mmb_fsync(struct file *file, struct mapping_metadata_bhs *mmb,
-		      loff_t start, loff_t end, bool datasync);
 void clean_bdev_aliases(struct block_device *bdev, sector_t block,
 			sector_t len);
 static inline void clean_bdev_bh_alias(struct buffer_head *bh)
diff --git a/include/linux/fs.h b/include/linux/fs.h
index caa9203ed213..32178e53d448 100644
--- a/include/linux/fs.h
+++ b/include/linux/fs.h
@@ -3298,8 +3298,16 @@ void simple_offset_destroy(struct offset_ctx *octx);
 
 extern const struct file_operations simple_offset_dir_operations;
 
-extern int __generic_file_fsync(struct file *, loff_t, loff_t, int);
-extern int generic_file_fsync(struct file *, loff_t, loff_t, int);
+int generic_mmb_fsync_noflush(struct file *file,
+			      struct mapping_metadata_bhs *mmb,
+			      loff_t start, loff_t end, bool datasync);
+int generic_mmb_fsync(struct file *file, struct mapping_metadata_bhs *mmb,
+		      loff_t start, loff_t end, bool datasync);
+static inline int generic_file_fsync(struct file *file,
+				     loff_t start, loff_t end, int datasync)
+{
+	return generic_mmb_fsync(file, NULL, start, end, datasync);
+}
 
 extern int generic_check_addressable(unsigned, u64);
 
diff --git a/fs/libfs.c b/fs/libfs.c
index 548e119668df..7c1d78862e39 100644
--- a/fs/libfs.c
+++ b/fs/libfs.c
@@ -18,7 +18,7 @@
 #include <linux/exportfs.h>
 #include <linux/iversion.h>
 #include <linux/writeback.h>
-#include <linux/buffer_head.h> /* sync_mapping_buffers */
+#include <linux/buffer_head.h> /* mmb_sync_buffers() */
 #include <linux/fs_context.h>
 #include <linux/pseudo_fs.h>
 #include <linux/fsnotify.h>
@@ -1539,19 +1539,22 @@ struct dentry *generic_fh_to_parent(struct super_block *sb, struct fid *fid,
 EXPORT_SYMBOL_GPL(generic_fh_to_parent);
 
 /**
- * __generic_file_fsync - generic fsync implementation for simple filesystems
+ * generic_mmb_fsync_noflush - generic buffer fsync implementation
+ * for simple filesystems with no inode lock
  *
  * @file:	file to synchronize
  * @start:	start offset in bytes
  * @end:	end offset in bytes (inclusive)
  * @datasync:	only synchronize essential metadata if true
  *
  * This is a generic implementation of the fsync method for simple
  * filesystems which track all non-inode metadata in the buffers list
  * hanging off the address_space structure.
  */
-int __generic_file_fsync(struct file *file, loff_t start, loff_t end,
-				 int datasync)
+int generic_mmb_fsync_noflush(struct file *file,
+			      struct mapping_metadata_bhs *mmb,
+			      loff_t start, loff_t end, bool datasync)
 {
 	struct inode *inode = file->f_mapping->host;
 	int err;
@@ -1561,45 +1564,53 @@ int __generic_file_fsync(struct file *file, loff_t start, loff_t end,
 	if (err)
 		return err;
 
-	inode_lock(inode);
+	if (mmb)
+		ret = mmb_sync_buffers(mmb);
 	if (!(inode_state_read_once(inode) & I_DIRTY_ALL))
 		goto out;
 	if (datasync && !(inode_state_read_once(inode) & I_DIRTY_DATASYNC))
 		goto out;
 
-	ret = sync_inode_metadata(inode, 1);
+	err = sync_inode_metadata(inode, 1);
+	if (ret == 0)
+		ret = err;
+
 out:
-	inode_unlock(inode);
 	/* check and advance again to catch errors after syncing out buffers */
 	err = file_check_and_advance_wb_err(file);
 	if (ret == 0)
 		ret = err;
 	return ret;
 }
-EXPORT_SYMBOL(__generic_file_fsync);
+EXPORT_SYMBOL(generic_mmb_fsync_noflush);
 
 /**
  * generic_file_fsync - generic fsync implementation for simple filesystems
  *			with flush
  * @file:	file to synchronize
+ * @mmb:	list of metadata bhs to flush
  * @start:	start offset in bytes
  * @end:	end offset in bytes (inclusive)
  * @datasync:	only synchronize essential metadata if true
  *
+ * This is a generic implementation of the fsync method for simple
+ * filesystems which track all non-inode metadata in the buffers list
+ * hanging off the address_space structure. This also makes sure that
+ * a device cache flush operation is called at the end.
  */
-
-int generic_file_fsync(struct file *file, loff_t start, loff_t end,
-		       int datasync)
+int generic_mmb_fsync(struct file *file, struct mapping_metadata_bhs *mmb,
+		      loff_t start, loff_t end, bool datasync)
 {
 	struct inode *inode = file->f_mapping->host;
-	int err;
+	int ret;
 
-	err = __generic_file_fsync(file, start, end, datasync);
-	if (err)
-		return err;
-	return blkdev_issue_flush(inode->i_sb->s_bdev);
+	ret = generic_mmb_fsync_noflush(file, mmb, start, end, datasync);
+	if (!ret)
+		ret = blkdev_issue_flush(inode->i_sb->s_bdev);
+	return ret;
 }
-EXPORT_SYMBOL(generic_file_fsync);
+EXPORT_SYMBOL(generic_mmb_fsync);
 
 /**
  * generic_check_addressable - Check addressability of file system
@@ -1539,19 +1539,22 @@ struct dentry *generic_fh_to_parent(struct super_block *sb, struct fid *fid,
  *
- * @file:	file to synchronize
- * @start:	start offset in bytes
- * @end:	end offset in bytes (inclusive)
- * @datasync:	only synchronize essential metadata if true
+ * @file:       file to synchronize
+ * @mmb:        list of metadata bhs to flush
+ * @start:      start offset in bytes
+ * @end:        end offset in bytes (inclusive)
+ * @datasync:   only synchronize essential metadata if true
  *
  * This is a generic implementation of the fsync method for simple
@@ -1561,45 +1564,53 @@ int __generic_file_fsync(struct file *file, loff_t start, loff_t end,
 
 /**
- * generic_file_fsync - generic fsync implementation for simple filesystems
- *			with flush
+ * generic_mmb_fsync - generic buffer fsync implementation
+ * for simple filesystems with no inode lock
+ *
  * @file:	file to synchronize
-- 
2.51.0


  parent reply	other threads:[~2026-03-20 13:44 UTC|newest]

Thread overview: 68+ messages / expand[flat|nested]  mbox.gz  Atom feed  top
2026-03-20 13:40 [PATCH v2 0/41] fs: Move metadata bh tracking from address_space Jan Kara
2026-03-20 13:40 ` [PATCH 01/41] ext4: Use inode_has_buffers() Jan Kara
2026-03-20 13:40 ` [PATCH 02/41] gfs2: Don't zero i_private_data Jan Kara
2026-03-20 13:40 ` [PATCH 03/41] ntfs3: Drop pointless sync_mapping_buffers() and invalidate_inode_buffers() calls Jan Kara
2026-03-20 13:40 ` [PATCH 04/41] ocfs2: Drop pointless sync_mapping_buffers() calls Jan Kara
2026-03-23 10:46   ` Joseph Qi
2026-03-20 13:41 ` [PATCH 05/41] bdev: Drop pointless invalidate_inode_buffers() call Jan Kara
2026-03-20 13:41 ` [PATCH 06/41] ufs: Drop pointless invalidate_mapping_buffers() call Jan Kara
2026-03-20 13:41 ` [PATCH 07/41] exfat: Drop pointless invalidate_inode_buffers() call Jan Kara
2026-03-20 13:41 ` [PATCH 08/41] udf: Switch to generic_buffers_fsync() Jan Kara
2026-03-24  5:38   ` Christoph Hellwig
2026-03-24 12:24     ` Jan Kara
2026-03-20 13:41 ` [PATCH 09/41] minix: " Jan Kara
2026-03-20 13:41 ` [PATCH 10/41] bfs: " Jan Kara
2026-03-20 13:41 ` [PATCH 11/41] fat: Switch to generic_buffers_fsync_noflush() Jan Kara
2026-03-20 13:41 ` [PATCH 12/41] fs: Drop sync_mapping_buffers() from __generic_file_fsync() Jan Kara
2026-03-24  5:40   ` Christoph Hellwig
2026-03-24 12:34     ` Jan Kara
2026-03-24 13:17       ` Christoph Hellwig
2026-03-24 13:36         ` Jan Kara
2026-03-24 15:54           ` Christoph Hellwig
2026-03-25 19:01             ` Jan Kara
2026-03-20 13:41 ` [PATCH 13/41] fat: Sync and invalidate metadata buffers from fat_evict_inode() Jan Kara
2026-03-20 13:41 ` [PATCH 14/41] udf: Sync and invalidate metadata buffers from udf_evict_inode() Jan Kara
2026-03-20 13:41 ` [PATCH 15/41] minix: Sync and invalidate metadata buffers from minix_evict_inode() Jan Kara
2026-03-20 13:41 ` [PATCH 16/41] ext2: Sync and invalidate metadata buffers from ext2_evict_inode() Jan Kara
2026-03-20 13:41 ` [PATCH 17/41] ext4: Sync and invalidate metadata buffers from ext4_evict_inode() Jan Kara
2026-03-20 13:41 ` [PATCH 18/41] bfs: Sync and invalidate metadata buffers from bfs_evict_inode() Jan Kara
2026-03-20 13:41 ` [PATCH 19/41] affs: Sync and invalidate metadata buffers from affs_evict_inode() Jan Kara
2026-03-20 13:41 ` [PATCH 20/41] fs: Ignore inode metadata buffers in inode_lru_isolate() Jan Kara
2026-03-24  5:42   ` Christoph Hellwig
2026-03-24 12:51     ` Jan Kara
2026-03-20 13:41 ` [PATCH 21/41] fs: Stop using i_private_data for metadata bh tracking Jan Kara
2026-03-24  5:42   ` Christoph Hellwig
2026-03-20 13:41 ` [PATCH 22/41] hugetlbfs: Stop using i_private_data Jan Kara
2026-03-24  5:42   ` Christoph Hellwig
2026-03-20 13:41 ` [PATCH 23/41] aio: Stop using i_private_data and i_private_lock Jan Kara
2026-03-24  5:43   ` Christoph Hellwig
2026-03-20 13:41 ` [PATCH 24/41] fs: Remove i_private_data Jan Kara
2026-03-24  5:43   ` Christoph Hellwig
2026-03-20 13:41 ` [PATCH 25/41] kvm: Use private inode list instead of i_private_list Jan Kara
2026-03-24  5:44   ` Christoph Hellwig
2026-03-20 13:41 ` [PATCH 26/41] fs: Drop osync_buffers_list() Jan Kara
2026-03-24  5:44   ` Christoph Hellwig
2026-03-20 13:41 ` [PATCH 27/41] fs: Fold fsync_buffers_list() into sync_mapping_buffers() Jan Kara
2026-03-24  5:44   ` Christoph Hellwig
2026-03-20 13:41 ` [PATCH 28/41] fs: Move metadata bhs tracking to a separate struct Jan Kara
2026-03-24  5:47   ` Christoph Hellwig
2026-03-20 13:41 ` [PATCH 29/41] fs: Make bhs point to mapping_metadata_bhs Jan Kara
2026-03-24  5:48   ` Christoph Hellwig
2026-03-20 13:41 ` [PATCH 30/41] fs: Switch inode_has_buffers() to take mapping_metadata_bhs Jan Kara
2026-03-24  5:48   ` Christoph Hellwig
2026-03-20 13:41 ` [PATCH 31/41] fs: Provide functions for handling mapping_metadata_bhs directly Jan Kara
2026-03-24  5:51   ` Christoph Hellwig
2026-03-25 19:00     ` Jan Kara
2026-03-20 13:41 ` [PATCH 32/41] ext2: Track metadata bhs in fs-private inode part Jan Kara
2026-03-20 13:41 ` [PATCH 33/41] affs: " Jan Kara
2026-03-20 13:41 ` [PATCH 34/41] bfs: " Jan Kara
2026-03-20 13:41 ` [PATCH 35/41] fat: " Jan Kara
2026-03-20 13:41 ` [PATCH 36/41] udf: " Jan Kara
2026-03-20 13:41 ` [PATCH 37/41] minix: " Jan Kara
2026-03-20 13:41 ` [PATCH 38/41] ext4: " Jan Kara
2026-03-20 13:41 ` [PATCH 39/41] fs: Drop mapping_metadata_bhs from address space Jan Kara
2026-03-20 13:41 ` [PATCH 40/41] fs: Drop i_private_list from address_space Jan Kara
2026-03-20 13:41 ` Jan Kara [this message]
2026-03-24  5:56   ` [PATCH 41/41] fs: Unify generic_file_fsync() with mmb methods Christoph Hellwig
2026-03-24 13:28     ` Jan Kara
2026-03-23 10:20 ` [PATCH v2 0/41] fs: Move metadata bh tracking from address_space Christian Brauner

Reply instructions:

You may reply publicly to this message via plain-text email
using any one of the following methods:

* Save the following mbox file, import it into your mail client,
  and reply-to-all from there: mbox

  Avoid top-posting and favor interleaved quoting:
  https://en.wikipedia.org/wiki/Posting_style#Interleaved_style

* Reply using the --to, --cc, and --in-reply-to
  switches of git-send-email(1):

  git send-email \
    --in-reply-to=20260320134100.20731-82-jack@suse.cz \
    --to=jack@suse.cz \
    --cc=aivazian.tigran@gmail.com \
    --cc=bcrl@kvack.org \
    --cc=brauner@kernel.org \
    --cc=david@kernel.org \
    --cc=dsterba@suse.com \
    --cc=hirofumi@mail.parknet.co.jp \
    --cc=linux-aio@kvack.org \
    --cc=linux-block@vger.kernel.org \
    --cc=linux-ext4@vger.kernel.org \
    --cc=linux-fsdevel@vger.kernel.org \
    --cc=linux-mm@kvack.org \
    --cc=muchun.song@linux.dev \
    --cc=osalvador@suse.de \
    --cc=tytso@mit.edu \
    --cc=viro@ZenIV.linux.org.uk \
    /path/to/YOUR_REPLY

  https://kernel.org/pub/software/scm/git/docs/git-send-email.html

* If your mail client supports setting the In-Reply-To header
  via mailto: links, try the mailto: link
Be sure your reply has a Subject: header at the top and a blank line before the message body.
This is a public inbox, see mirroring instructions
for how to clone and mirror all data and code used for this inbox